require 'digest'
require 'json'


class KeyValueChain
  def initialize(path)
    @path = path
    @chain = DriveKVChain.new(path)
  end
  def [](key)
    @chain[key]
  end
  def []=(key,value)
    @chain[key] = value
  end

  class KVPair
    attr_reader :key, :value
    def initialize(prevBlockHash, key, value, nonce = -1)
      @prevBlockHash = prevBlockHash
      @key = key
      @value = value
      @nonce = nonce
      if @nonce < 0
        @none = 0
        until self.is_valid?
          @nonce += 1
        end
      end
    end
    def to_hash
      Digest::SHA1.hexdigest self.to_json
    end
    def from_json(json)
      h = JSON.parse(json)
      @prevBlockHash = h["prevBlock"]
      @nonce = h["nonce"]
      @key = h["key"]
      @value = h["value"]
    end
    def to_json
      pairhash = {:prevBlock => @prevBlockHash, :nonce => @nonce, :key => @key, :value => @value}
      JSON.generate(pairhash)
    end
    def is_valid?
      self.to_hash[0,4] == "0000"
    end
  end

  class DriveKVChain
    def initialize(path)
      @path = path
      @last_block_hash = "0000000000000000000000000000000000000000"
      self[""]
    end
    def [](key)
      prev = "0000000000000000000000000000000000000000"
      value = ""
      File.open(@path, "a+") do |f|
        f.each_line do |line|
          h = JSON.parse(line)
          raise "Broken Blockchain, invalid hashes" unless h["prevBlock"] == prev
          prev = Digest::SHA1.hexdigest line.chomp
          raise "Broken Blockchain, invalid nonce" unless prev[0,4] == "0000"
          value = h["value"] if h["key"] == key
        end
      end
      @last_block_hash = prev
      value
    end

    def []=(key,value)
      pair = KVPair.new(@last_block_hash, key, value)
      open(@path, 'a') { |f|
        f.puts pair.to_json
      }
      @last_block_hash = pair.to_hash
    end
  end

  class MemKVChain
    def initialize(path)
      @blocks = []
      @hashtable = Hash.new("")
      @path = path
      self.loadfile
    end
    def add_block_to_hash(pair)
      if pair.is_valid?
        @blocks << pair
        @hashtable[pair.key] = pair.value
        return true
      end
      false
    end

    def add_block(pair)
      if add_block_to_hash(pair)
        open(@path, 'a') { |f|
          f.puts pair.to_json
        }
        return true
      else
        return false
      end
    end
    def last_block_json
      @blocks[-1].to_json
    end
    def last_block_hash
      if @blocks.any?
        @blocks[-1].to_hash
      else
        "0000000000000000000000000000000000000000"
      end
    end
    def [](key)
      @hashtable[key]
    end
    def []=(key, value)
      self.add_block(KVPair.new(self.last_block_hash, key, value))
    end
    def loadfile(path = @path)
      #f = File.open(path, "a+")
      #f.foreach(path).with_index do |line, line_num|
      #  self.add_block(pair_from_json(line))
      #end
      File.open(path, "a+") do |f|
        f.each_line do |line|
          self.add_block_to_hash(pair_from_json(line))
        end
      end

    end
  end

  def pair_from_json(json)
    pair = KVPair.new("", "", "", 0)
    pair.from_json(json)
    pair
  end

end





#chain = DriveKVChain.new("./logfile")
#(1..1000).each { |n| puts chain["key#{n}"]}
